

// FunctionalIIC v.0.2. 20210705. by ken woo.
// the master, not only takes care of negotiating to slaves but also the backend data of in/out processing as well as at realtime.
// so, data can be not yet ready or buffer full and SCL waiting for available. both statical and dynamical data processing are support.
// dynamically of cases are, for examples, 1) master definitely writes, then reads some then accordingly decides how many to totally read.
// 2) master is writing the data while is slowly generating by backend and has to wait. and so on.
// also supports the successive requests to the master nonblockingly, for example,
// read slave A and write slave B and write slave A and read slave C by 4 cIIC objects successively invoked in main function.
// usage: static mode, the data read/write buffers and the lengths to read/write must be determined.
// dynamic mode, the _RT in code, 6 parameters. 2 are read-cb and write-cb functions. 2 are read/write buffers. 2 are read/write lengths.
// lengths are nonzero if precedently known about at least how many to read/write; so are the buffers accordingly to be determined.
// however each of the buffers and lengths can be lately determined in cb functions. (read buff?, :).
// the cb funcs, write-cb is called before master doing the writing-data stage, for user preparing data realtime.
// so is the read-cb though, read buffer is determined in advance, except length 0.
// in addition, can be changed inside the read-cb function after the data read and the fulfilled buffer; can then to swap buffer.
// the cb functions defined by user have to be IRAM_ATTR if timer-triggered isr-func used.
// however, isr-func can be polling without using timer-triggered method where probably better bit-rate than timer triggered.

// AAA -------- platform dependent parameters -----------
#define MY_IIC_DEBUG                1 // only for logistic look output up. enable could make slaves timeout and drop transfer. must disabled.
#define SRAM_ACES                   IRAM_ATTR
#define GPIO_TR_TIME_US             1
#define MYDELAY_US(x)               delayMicroseconds(x)
#define IIC_BUS_DELAY               MYDELAY_US(GPIO_TR_TIME_US)
#define SCLIO                       5
#define SDAIO                       4
#define SCL_H                       pinMode(SCLIO, INPUT_PULLUP)/////{pinMode(SCLIO, OUTPUT); digitalWrite(SCLIO, 1);}
#define SDA_H                       pinMode(SDAIO, INPUT_PULLUP)/////{pinMode(SDAIO, OUTPUT); digitalWrite(SDAIO, 1);}
#define SCL_L                       {pinMode(SCLIO, OUTPUT); digitalWrite(SCLIO, 0);}
#define SDA_L                       {pinMode(SDAIO, OUTPUT); digitalWrite(SDAIO, 0);}

unsigned digitalReadOutputPin(unsigned pin){ // https://forum.arduino.cc/t/how-to-find-out-output-pin-state/42066
    unsigned abit = digitalPinToBitMask(pin);
    unsigned aport = digitalPinToPort(pin);

    if (!(*portModeRegister(aport)&abit)) return digitalRead(pin); // (1==output, 0==input).

    if (aport == NOT_A_PIN) return 0;
    return (*portOutputRegister(aport) & abit) ? 1 : 0;
}

#define IS_SCL_H                    digitalReadOutputPin(SCLIO)
#define IS_SDA_H                    digitalReadOutputPin(SDAIO)
// VVV -------- platform dependent parameters -----------

#if (MY_IIC_DEBUG)
    #define MYIICDEBUG(x)           (x)

    const char* dbg_iic_stage_name(unsigned proc){
        static const char *const z[]={
            "IDLE",
            "START",
            "REP-START",
            "RETRY",
            "ADDR",
            "WRITE",
            "READ",
            "WAIT",
            "STOP",
            "RESET"
        };
        return z[proc];
    }
#else
    #define MYIICDEBUG(x)
#endif




typedef class __attribute__((packed, aligned(4))) cIIC{

    #define IIC_TIMER_PERIOD        34 // must >= 34us; any time can be for every request.
    #define IIC_TMO_BYTE_RETRY      2 // default 2 times plus org one totally 3 times. set 0 no retry.
    #define IIC_TMO_ADDR_RETRY      2 // default 2 times plus org one totally 3 times. set 0 no retry.
    #define IIC_TMO_WAIT_ADDRESS    8 // wait for 8 timer-cycle for scl bus becoming normal under address stage; must >=1.
    #define IIC_TMO_WAIT_CLK_RW     2 // wait for 2 timer-cycle for scl bus becoming normal under data stage; must >=1.
    #define IIC_TMO_RESET_CNT       16 // default send 16 timer-cycle scl-toggling which is 8 scl clock cycles; must >=16 and even number.
    #define IIC_TMO_WAIT_RTDATA(x)  ((x)-1) // default 3, must >=2. timer cycles to drop this transfer if timeout and no RT data.

    typedef enum eIicProc{
        IIC_PROC_IDLE=              0,
        IIC_PROC_START=             1,
        IIC_PROC_REPEATSTART=       2,
        IIC_PROC_RETRY=             3,
        IIC_PROC_ADDRESS=           4,
        IIC_PROC_WRITE_DATA=        5,
        IIC_PROC_READ_DATA=         6,
        IIC_PROC_WAIT_STATES=       7,
        IIC_PROC_STOP=              8,
        IIC_PROC_RESET=             9,
    } eIicProc;

    typedef enum eIicFlags{
        IIC_FLAG_IDLE=              0x00,
        IIC_FLAG_BUSY=              0x01,
        IIC_FLAG_NACK=              0x02,
        IIC_FLAG_RESET=             0x04,
        IIC_FLAG_SUCCESS=           0x08,
        IIC_FLAG_WAIT_CLK_H_ADDR=   0x10,
        IIC_FLAG_WAIT_CLK_H_DATA=   0x20,
        IIC_FLAG_WAIT_DATA=         0x40,
        IIC_FLAG_WAIT_STATES=       (0x70),
        IIC_FLAG_STATUS=            (0x7F),
        IIC_FLAG_WRITE=             0x80,
        IIC_FLAG_READ=              0x100,
        IIC_FLAG_WRITE_FIRST=       0x200,
        IIC_FLAG_TRANSFER=          (0x380),
        IIC_FLAG_RT_WRITE=          0x400,
        IIC_FLAG_RT_READ=           0x800,
        IIC_FLAG_RT_TRANSFER=       (0xC00),
        IIC_FLAG_SR_CONTINUE=       0x1000,
        IIC_FLAG_SR_RETRY=          0x2000,
        IIC_FLAG_SR_NEW=            0x4000,
        IIC_FLAG_SR_STATES=         (0x7000),
        IIC_FLAG_DIRECT_STOP=       0x8000,
    } eIicFlags;

    typedef struct __attribute__((packed, aligned(4))) sIicMasterObj{
        unsigned address;           // 7-bit slave address
        unsigned read_cnt;          // the number of bytes to master read
        unsigned write_cnt;         // the number of bytes to master write
        unsigned char *rt_ibuff;    // the put-to data array pointer for master read
        unsigned char *rt_obuff;    // the get-from data array pointer for master write
        void (*sr_s_cb_func)();     // this cb_func called when this request is starting at start/sr stage
        void (*sr_p_cb_func)();     // this cb_func called when this request is done at sr/stop stage
        void (*rt_wdata_prepare)(); // preparing the realtime data of output, cb_func
        void (*rt_rdata_postacc)(); // the realtime data by input for post access, cb_func
        unsigned flags;             // for statical and dynamical states, from eIicFlags
        void Clear(){
            int i=sizeof(sIicMasterObj)/sizeof(unsigned);
            while (i--) (reinterpret_cast<unsigned*>(&(*this)))[i]=0;
        };
        sIicMasterObj(){Clear();};
    } sIicMasterObj;

    typedef struct __attribute__((packed, aligned(4))) sIicQueue{ // all the queue objs must be newed.
        sIicMasterObj *obj;
        sIicQueue *nxt;
        sIicQueue(): obj(0), nxt(0){;};

        void Add(sIicQueue **predecessor_pointer, sIicMasterObj *masterptr){
            // add to tail. and since it is adding to tail where would not do by isr, so no need gate control.
            // note, there must be a predecessor.

            if (!predecessor_pointer) return;
            sIicQueue *a=new sIicQueue;
            a->obj=masterptr;
            while (*predecessor_pointer) predecessor_pointer=&((*predecessor_pointer)->nxt);
            *predecessor_pointer=a;
        };

        SRAM_ACES void Remove(sIicQueue **possessor_pointer, sIicQueue **new_por=0){
            // remove head(or any node) and collect into new_por. note, there must be a possessor.
            // if new_por==0, directly delete the node, else gc for later treatment; @config() is best.

            if (/*!this || */!possessor_pointer || !*possessor_pointer) return;
            sIicQueue *next=*possessor_pointer;
            if (new_por){ // for GarbageCollection
                next=next->nxt;
                (*possessor_pointer)->nxt=*new_por;
                *new_por=*possessor_pointer;
                *possessor_pointer=next;
            }
            else {
                *possessor_pointer=next->nxt;
                next->nxt=0;
                delete next;
            }
        };

        ~sIicQueue(){if (this && this->nxt) delete this->nxt;}; // remove entire list.
    } sIicQueue;

    typedef struct __attribute__((packed, aligned(4))) sIicMasterSig{ // for master signaling. it must be inited before all iic objects running.
        unsigned the_rt_all_bytes;  // designate static mode(1), total data length is still indeterminate(0).
        unsigned the_bytes;         // how many data bytes to transfer
        unsigned char **the_buff;   // the current accessing data array. the ptr of ptr so can change buffer at runtime.
        void (*the_data_cb)();      // the current realtime data processing cb-function.
        unsigned the_byte_loc;      // the current accessing byte location
        unsigned the_data;          // the current accessing data byte
        unsigned the_bit_loc;       // the current bit location of the accessed data byte
        void (**the_procs)();       // the current proc-call-func located by the_proc
        eIicProc the_proc;          // the current controlling state, as the index of the_procs
        eIicProc the_next_proc;     // the next controlling state for return after exceptional events handled
        unsigned the_flags;         // the current states
        sIicQueue *the_queue;       // for queueing requests of data transfer
        sIicQueue *the_unqueue;     // for garbage collection
        sIicQueue *the_gate;        // prevent from data access violation; altered by MasterObj, checked by sIicMasterSig. no hang on value.
        unsigned the_clk_flip;      // a timer trigger is a half cycle, so need flip
        unsigned the_retry_cnt;     // for retry counter
        unsigned the_tmo_cnt;       // for timeout counter
        unsigned temp;              // yes, better a temp keeps trace, used only in procedural functions since mutually exclusive.
        void Clear(){
            int i=sizeof(sIicMasterSig)/sizeof(unsigned);
            while (i--) (reinterpret_cast<unsigned*>(&(*this)))[i]=0;
        };
        sIicMasterSig(){Clear();};
    } sIicMasterSig;

    SRAM_ACES static sIicMasterSig sig;
    sIicMasterObj master;

    SRAM_ACES static void SdaLow(){SDA_L;};
    SRAM_ACES static void SdaLow_w(){SDA_L; IIC_BUS_DELAY;}; // _w, wait for a gpio transition time, post delay.
    SRAM_ACES static void SdaLow_W(){IIC_BUS_DELAY; SDA_L;}; // _W, wait for a gpio transition time, prefix delay.
    SRAM_ACES static void SdaHigh(){SDA_H;};
    SRAM_ACES static void SdaHigh_w(){SDA_H; IIC_BUS_DELAY;};
    SRAM_ACES static void SdaHigh_W(){IIC_BUS_DELAY; SDA_H;};
    SRAM_ACES static void SclLow(){SCL_L;};
    SRAM_ACES static void SclLow_w(){SCL_L; IIC_BUS_DELAY;};
    SRAM_ACES static void SclLow_W(){IIC_BUS_DELAY; SCL_L;};
    SRAM_ACES static void SclHigh(){SCL_H;};
    SRAM_ACES static void SclHigh_w(){SCL_H; IIC_BUS_DELAY;};
    SRAM_ACES static void SclHigh_W(){IIC_BUS_DELAY; SCL_H;};

    SRAM_ACES static unsigned isSdaLow(){return !IS_SDA_H;};
    SRAM_ACES static unsigned isSdaHigh(){return IS_SDA_H;};
    SRAM_ACES static unsigned isSclLow(){return !IS_SCL_H;};
    SRAM_ACES static unsigned isSclLow_W(){IIC_BUS_DELAY; return !IS_SCL_H;};
    SRAM_ACES static unsigned isSclHigh(){return IS_SCL_H;};

    SRAM_ACES static void iicproc_reset(){
        // attempt to send START byte 0x01=0000000+R(1). and the scl toggling is the key. only set cnt 16 can match.
        // and then go to stop.
        // in(SCLl, in any case).
        // out0(SCLD2h, SDAh, IIC_PROC_STOP).

        if (!sig.temp){ // begin to toggling.
            sig.temp=(IIC_TMO_RESET_CNT-1);
            sig.the_clk_flip=0;
            SclLow_w();
            SdaLow();
            return;
        }
        else if (--sig.temp){
            sig.the_clk_flip^=1;
            if (sig.the_clk_flip) SclHigh();
            else SclLow();
            if (sig.temp==1) SdaHigh_W(); // the last bit.
        }
        else { // terminated. @C2.
            SclHigh();
            sig.the_flags|=IIC_FLAG_RESET;
            sig.the_proc=IIC_PROC_STOP;
        }
    };

    SRAM_ACES static void iicproc_wait_states(){
        // in(SCLD1l, SDAx) if SCLl then to wait. from address stage.
        // in(SCLDil, SDAx) if SCLl then to wait, waiting for slave preparing data.
        // in(SCLB1l, SDAx) if SCLl and waiting for master preparing data, read phase.
        // in(SCLA0h, SDAx) waiting for master preparing data, write phase.
        // out(SCLh, no change) if clk-wait success.
        // out(SCLl, IIC_PROC_RESET) if failed.
        // out(SCLB1l, no change) if master preparing done or timeout.

        if (!sig.temp) sig.temp=(sig.the_tmo_cnt-1); // to begin wait.
        else sig.temp--;
        switch (sig.the_flags&IIC_FLAG_WAIT_STATES){
            case IIC_FLAG_WAIT_CLK_H_ADDR:
                if (isSclHigh()){ // finally
                    sig.the_flags&=~IIC_FLAG_WAIT_CLK_H_ADDR;
                    sig.the_proc=sig.the_next_proc;
                    sig.temp=0;
                }
                else if (sig.temp) return; // wait more.
                else { // timeout and scl bus failed.
                    sig.the_proc=IIC_PROC_RESET;
                }
                break;
            case IIC_FLAG_WAIT_CLK_H_DATA:
                if (isSclHigh()){ // finally
                    sig.the_flags&=~IIC_FLAG_WAIT_CLK_H_DATA;
                    sig.the_proc=sig.the_next_proc;
                    sig.temp=0;

                    sig.the_clk_flip=1; // clock now high.
                }
                else if (sig.temp) return; // wait more.
                else { // timeout and scl bus failed.
                    sig.the_proc=IIC_PROC_RESET;
                }
                break;
            case IIC_FLAG_WAIT_DATA:
                (sig.the_data_cb)();
                if (sig.the_byte_loc==sig.the_bytes){ // means remains none.
                    if (!sig.temp){ // timeout
                        sig.the_rt_all_bytes=1; // let it back and branch to nack slave.
                        sig.the_proc=sig.the_next_proc;
                    }
                }
                else {
                    sig.temp=0;
                    sig.the_flags&=~IIC_FLAG_WAIT_DATA;
                    sig.the_proc=sig.the_next_proc;
                }
                break;
        }
    };

    SRAM_ACES static void iicproc_retry(){ // the next cycle, do nothing.
    };

    SRAM_ACES static void iicproc_repeatstart(){
        // in(SCLB0l, SDAh) if retry by nacked.
        // in(SCLB0l, SDAh) if new, continue.
        // out(SCLA9h, SDAl, flip1, b9, sig.the_next_proc).

        switch (sig.the_flags&IIC_FLAG_SR_STATES){
            case IIC_FLAG_SR_RETRY: // in the retry state, for address.
            case IIC_FLAG_SR_CONTINUE: // during a r-to-w or w-to-r session.
                SclHigh_w();
                sig.the_data=(sig.the_queue->obj->address|((sig.the_flags&IIC_FLAG_READ)? 1: 0));
                break;
            case IIC_FLAG_SR_NEW: // the new request to another slave
                if (sig.the_gate) return;
                SclHigh_w();

                    MYIICDEBUG(printf("\r\n<<<result flag: 0x%x>>>\r\n", sig.the_flags));

                sig.the_queue->obj->flags&=~IIC_FLAG_STATUS;
                sig.the_queue->obj->flags|=(sig.the_flags&IIC_FLAG_STATUS);

                if (sig.the_queue->obj->sr_p_cb_func) sig.the_queue->obj->sr_p_cb_func();

                sig.the_queue->Remove(&(sig.the_queue), &(sig.the_unqueue));

                if (sig.the_queue->obj->sr_s_cb_func) sig.the_queue->obj->sr_s_cb_func();

                sig.the_flags=(sig.the_queue->obj->flags|IIC_FLAG_BUSY);
                sig.the_data=(sig.the_queue->obj->address|(((sig.the_flags&IIC_FLAG_WRITE_FIRST)==IIC_FLAG_WRITE_FIRST)? 0: 1));
                break;
        }

        SdaLow_W();

        sig.the_flags&=~IIC_FLAG_SR_STATES;
        sig.the_bit_loc=9;
        sig.the_clk_flip=1; // clock now high.
        sig.the_proc=sig.the_next_proc;
    };

    SRAM_ACES static void iicproc_start(){
        // in(SCLA9h, SDAh).
        // out(SCLA9h, SDAl, flip1, b9).
        //
        // note the designation:
        // start bit(SCL_h, SDA_h_to_l). the SCL_h then, \_/- are ABCD respectively.
        // again, the 1st bit after start bit is A9_h_to_l(\), B9_keep_low(_), C9_l_to_h(/), D9_keep_high.
        // yet again, 9 to 2 are address or data bits, 1 is ack/nack bit. 0 is the succeeding bit,
        // which will be A0B0C0D0 equals to A9B9C9D9 of a read-bit/write-bit cycle(case1),
        // or equals to AsBsCsDs of a start/repeatstart cycle(case2),
        // or equals to ApBpCpDp of a stop cycle(case3).
        // for SDA_change, which is SDAc(x),
        // for case1: SDAc(after A, or say B) for write, SDAc(at the end of D, or say before next A) for read. and so do following bits.
        // for case2: SDAc(after A0 if SDA_l) and SDAc(at the begin of D0, or say during D0) for both start/repeatstart.
        // for case3: SDAc(after A0 if SDA_h) and SDAc(during D0, or say after C0 but a bit wait) for stop.
        // so for simplicity, the case1, the write processed at after A(h-to-l) the read at before A(h-to-l).

        if (sig.the_queue->obj->sr_s_cb_func) sig.the_queue->obj->sr_s_cb_func();

        SdaLow();

        sig.the_flags=(sig.the_queue->obj->flags|IIC_FLAG_BUSY);
        sig.the_data=(sig.the_queue->obj->address|(((sig.the_flags&IIC_FLAG_WRITE_FIRST)==IIC_FLAG_WRITE_FIRST)? 0: 1));
        sig.the_bit_loc=9;
        sig.the_clk_flip=1; // clock now high.
        sig.the_proc=IIC_PROC_ADDRESS;
    };

    SRAM_ACES static void iicproc_address(){ // byte level bus check.
        // in(SCLA9h, SDAl, flip1, b9) start, repeat start.
        // out(SCLA0h, SDAl) if acked then write.
        // out(SCLB0l, SDAh) if acked then read.
        // out(SCLD1l, SDAx) if should be SCLh but SCLl then to wait.
        // out(SCLB0l, SDAh) if nacked then to retry.
        // out(SCLB0l, SDAl) if nacked then to stop.

        if (sig.the_clk_flip^=1){ // to high level.
            SclHigh(); // @C9 to C1
            if (!sig.the_bit_loc && isSclLow_W()){ // @C1_low_to_high_ack state, then got D1_low.
                // case might exist that if the slave addressed who slow to prepare data,
                // may pull scl bus low for wait state during ack stage. so needs to check if happen and wait.
                // and usually, such a condition probably wait for longer.
                // must return into new stage.
                sig.the_flags|=IIC_FLAG_WAIT_CLK_H_ADDR;
                sig.the_proc=IIC_PROC_WAIT_STATES;
                sig.the_next_proc=IIC_PROC_ADDRESS;
                sig.the_tmo_cnt=IIC_TMO_WAIT_ADDRESS;
            }
        }
        else { // to low level. @A.
            if (sig.the_bit_loc){
                SclLow(); // B9 to B1
                if (--sig.the_bit_loc){ // B9 to B2
                    if (sig.the_data&(1<<(sig.the_bit_loc-1))) SdaHigh();
                    else SdaLow();
                }
                else SdaHigh(); // data bus to idle for slave ack. @B1.
            }
            else { // end address phase, @A0_scl_h_to_l.
                if (isSdaLow()){ // acked by slave, prepare to be SCL=B0low, SDA=B0low.
                    SdaLow(); // pull sda low by master for next data preparation, keep intact first.

                    sig.the_retry_cnt=0; // since it is used in the retry else-block.

                    // the read/write processes.
                    //
                    // note that the IIC_FLAG_WRITE_FIRST could decide 1st stage,
                    // the IIC_FLAG_WRITE/READ decides the 2nd stage whether or not exists;
                    // so have to clear bit here and only here to avoid repeated; elsewhere would clear bit in order to cancel succeedings.
                    if ((sig.the_flags&IIC_FLAG_WRITE_FIRST)==IIC_FLAG_WRITE_FIRST){ // out(SCL=A0high, SDA=low).
                        sig.the_flags&=~IIC_FLAG_WRITE_FIRST;
                        sig.the_flags&=~IIC_FLAG_WRITE;
                        sig.the_proc=IIC_PROC_WRITE_DATA;
                        if (sig.the_flags&IIC_FLAG_RT_WRITE){
                            sig.the_data_cb=sig.the_queue->obj->rt_wdata_prepare;
                            sig.the_rt_all_bytes=0;
                        }
                        else sig.the_rt_all_bytes=1;

                        // note: the sensitive vars are buff, index, and read/write counts which could modified by user at runtime.
                        // however these are allowed to change only inside cb functions which are called by this isr.
                        // and it allows user to call cb functions but only to modify the total count in increase manner,
                        // which would not cause violation; since modify the role of buffer outside cb is more complicated and not support.
                        // so the cases of data violations are only to add/remove queue which need gate control.
                        sig.the_buff=&(sig.the_queue->obj->rt_obuff);
                        sig.the_bytes=sig.the_queue->obj->write_cnt;
                        sig.the_clk_flip=1; // now is high
                        sig.the_byte_loc=0; // should be modified by data-producer.
                        sig.the_bit_loc=0; //// (sig.the_bytes? 10: 0);
                    }
                    else if ((sig.the_flags&IIC_FLAG_READ)==IIC_FLAG_READ){ // out(SCL=B0low, SDA=high).
                        SclLow_w();
                        SdaHigh(); // release sda bus since for slave to put data
                        sig.the_flags&=~IIC_FLAG_READ;
                        sig.the_proc=IIC_PROC_READ_DATA;
                        if (sig.the_flags&IIC_FLAG_RT_READ){
                            sig.the_data_cb=sig.the_queue->obj->rt_rdata_postacc;
                            sig.the_rt_all_bytes=0;
                        }
                        else sig.the_rt_all_bytes=1;
                        sig.the_buff=&(sig.the_queue->obj->rt_ibuff);
                        sig.the_bytes=sig.the_queue->obj->read_cnt;
                        //// sig.the_clk_flip=0; // now is low
                        sig.the_data=0; // should be modified by data-consumer.
                        sig.the_byte_loc=0; // should be modified by data-consumer.
                        sig.the_bit_loc=9;
                    }
                    else if ((sig.the_flags&IIC_FLAG_WRITE)==IIC_FLAG_WRITE){ // out(SCL=A0high, SDA=low).
                        sig.the_flags&=~IIC_FLAG_WRITE;
                        sig.the_proc=IIC_PROC_WRITE_DATA;
                        if (sig.the_flags&IIC_FLAG_RT_WRITE){
                            sig.the_data_cb=sig.the_queue->obj->rt_wdata_prepare;
                            sig.the_rt_all_bytes=0;
                        }
                        else sig.the_rt_all_bytes=1;
                        sig.the_buff=&(sig.the_queue->obj->rt_obuff);
                        sig.the_bytes=sig.the_queue->obj->write_cnt;
                        sig.the_clk_flip=1; // now is high
                        sig.the_byte_loc=0; // should be modified by data-producer.
                        sig.the_bit_loc=0; //// (sig.the_bytes? 10: 0);
                    }
                    else while (1); // no such condition
                }
                else { // no slaves response, post state is SCL=B0low. SDA=low if stop, SDA=high if retry.
                    // retry for IIC_TMO_ADDR_RETRY times.
                    SclLow();
                    if (!IIC_TMO_ADDR_RETRY){
                        SdaLow();
                        sig.the_proc=IIC_PROC_STOP;
                        sig.the_flags=IIC_FLAG_NACK;
                    }
                    else if (sig.the_retry_cnt){
                        if (--sig.the_retry_cnt){
                            //// SdaHigh();
                            sig.the_proc=IIC_PROC_REPEATSTART;
                            sig.the_next_proc=IIC_PROC_ADDRESS;
                            sig.the_flags|=IIC_FLAG_SR_RETRY;
                        }
                        else {
                            SdaLow();
                            sig.the_proc=IIC_PROC_STOP;
                            sig.the_flags=IIC_FLAG_NACK;
                        }
                    }
                    else { // begin to retry.
                        //// SdaHigh();
                        sig.the_retry_cnt=IIC_TMO_ADDR_RETRY;
                        sig.the_proc=IIC_PROC_REPEATSTART;
                        sig.the_next_proc=IIC_PROC_ADDRESS;
                        sig.the_flags|=IIC_FLAG_SR_RETRY;
                    }
                }
            }
        }
    };

    SRAM_ACES static void iicproc_read_data(){
        // in(SCLB0l, SDAh, flip0, b9).
        // out(SCLB0l, SDAl) for stop.
        // out(SCLB1l, SDAx) for wait preparing data.
        // out(SCLB0l, SDAh) for repeat start: w-continue, new slave.

        if (sig.the_clk_flip){ // now high and already have data waiting. @D.
            sig.the_clk_flip=0;
            if (--sig.the_bit_loc){ // bit-level reading. @D9 to D2.
                if (isSdaHigh()) sig.the_data|=(1<<(sig.the_bit_loc-1));
                SclLow(); // to low
                return;
            }
            else { // ack stage. at high level and going to low for next data byte. @D1.
                SclLow(); // for next data byte. @A0.
                if (sig.the_byte_loc<sig.the_bytes){ // since only either < or ==.
                    SdaHigh(); // release sda bus. @B0.
                    sig.the_data=0;
                    sig.the_bit_loc=9;
                }
                else { // terminated condition. @B0.
                    if (sig.the_flags&IIC_FLAG_WAIT_DATA){ // the previous preparing read timeout.
                        // so need to drop the succeeding write if have.
                        sig.the_flags&=~IIC_FLAG_WRITE;
                    }
                    if (sig.the_flags&IIC_FLAG_WRITE){
                        //// SdaHigh(); // release sda bus.
                        sig.the_flags|=IIC_FLAG_SR_CONTINUE;
                        sig.the_proc=IIC_PROC_REPEATSTART;
                        sig.the_next_proc=IIC_PROC_ADDRESS;
                    }
                    else if (sig.the_queue->nxt && !(sig.the_flags&IIC_OPT_DIRECT_STOP)){
                        //// SdaHigh(); // release sda bus.
                        sig.the_flags|=IIC_FLAG_SR_NEW;
                        sig.the_proc=IIC_PROC_REPEATSTART;
                        sig.the_next_proc=IIC_PROC_ADDRESS;
                    }
                    else { // since previous is scl_h and sda_h, so we scl_l, sda_l for preparing stop.
                        SdaLow();
                        if (!(sig.the_flags&IIC_FLAG_WAIT_DATA)) sig.the_flags=IIC_FLAG_SUCCESS;
                        sig.the_proc=IIC_PROC_STOP;
                    }
                }
            }
        }
        else { // now low. from @C9 to C1.
            if (sig.the_bit_loc>=2){ // data bits
                SclHigh(); // C9 to C2
                if (isSclLow_W()){ // clock stretched by slave at bit level
                    sig.the_proc=IIC_PROC_WAIT_STATES;
                    sig.the_next_proc=IIC_PROC_READ_DATA;
                    sig.the_flags|=IIC_FLAG_WAIT_CLK_H_DATA;
                    sig.the_tmo_cnt=IIC_TMO_WAIT_CLK_RW;
                    return;
                }
                sig.the_clk_flip=1;
            }
            else { // ack stage. ack it at scl low to high level. @B1.
                if (sig.the_byte_loc<sig.the_bytes){
                    (*(sig.the_buff))[sig.the_byte_loc++]=sig.the_data;
                    if ((sig.the_byte_loc==sig.the_bytes) && sig.the_rt_all_bytes){ // terminated condition, nacked to slave.
                        SdaHigh_w(); // nack to slave. @B1.
                        SclHigh(); // @C1.
                        sig.the_clk_flip=1;
                        return;
                    }
                    SdaLow_w(); // ack to slave. @B1.
                    SclHigh(); // @C1.
                    sig.the_clk_flip=1;
                }
                else if (!sig.the_rt_all_bytes){ // and data preparing wait is here.
                    // do the clk wait while cb-func, the (sig.the_data_cb)().
                    // note that the cb function has 4 vars to modify:
                    // the_rt_all_bytes to 1 if no more read preparing,
                    // sig.the_byte_loc to 0 for a new start index to the new/old buffer, or keep intact.
                    // sig.the_bytes to the number of all bytes(corresponds to the_byte_loc) which is ready.
                    // sig.the_buff if need to change buffer target.

                    sig.the_flags|=IIC_FLAG_WAIT_DATA;
                    sig.the_proc=IIC_PROC_WAIT_STATES;
                    sig.the_next_proc=IIC_PROC_READ_DATA;
                    sig.the_tmo_cnt=IIC_TMO_WAIT_RTDATA(3);
                }
                else { // terminated condition, nacked to slave.
                    SdaHigh_w(); // nack to slave. @B1.
                    SclHigh(); // @C1.
                    sig.the_clk_flip=1;
                }
            }
        }
    };

    SRAM_ACES static void iicproc_write_data(){ // note in write, the numbering increase 1; from A9 to A0 become from A10 to A1.
        // in(SCLA0h, SDAl, flip1, b0).
        // out(SCLA0h, SDAh) for naked then retry.
        // out(SCLA0l, SDAh) for repeat start: w-continue, new slave.
        // out(SCLA0l, SDAl) for stop.
        // out(SCLA0h, SDAx) for wait preparing data. note that the high will let slave stay with. however it is high(scl released).

        if (sig.the_bit_loc){
            if (sig.the_clk_flip){ // data bit-level. A10 to A1.
                if (sig.the_bit_loc>=3){ // bit-level. A10 to A3.
                    SclLow(); // to low
                    sig.the_clk_flip=0;
                    if (sig.the_data&(1<<(--sig.the_bit_loc-2))) SdaHigh();
                    else SdaLow();
                    return;
                }
                else if (sig.the_bit_loc==2){ // release data bus. @A2h, pre-ack stage. high to low, now B2l.
                    SclLow(); // to low
                    sig.the_bit_loc=1;
                    sig.the_clk_flip=0;
                    SdaHigh();
                    return;
                }
                else { // A1.
                    sig.the_bit_loc=0;
                    if (isSdaHigh()){ // nacked by slave
                        if (IIC_TMO_BYTE_RETRY){
                            if (sig.the_retry_cnt){
                                if (--sig.the_retry_cnt){
                                    sig.the_proc=IIC_PROC_RETRY;
                                    sig.the_next_proc=IIC_PROC_WRITE_DATA;
                                    sig.the_bit_loc=10;
                                    return;
                                }
                                else {
                                    sig.the_flags|=IIC_FLAG_NACK;
                                    sig.the_flags&=~IIC_FLAG_READ;
                                    sig.the_byte_loc=sig.the_bytes;
                                    sig.the_rt_all_bytes=1;
                                }
                            }
                            else {
                                sig.the_retry_cnt=IIC_TMO_BYTE_RETRY;
                                sig.the_proc=IIC_PROC_RETRY;
                                sig.the_next_proc=IIC_PROC_WRITE_DATA;
                                sig.the_bit_loc=10;
                                return;
                            }
                        }
                        else { // no retry, terminate current master session.
                            sig.the_flags|=IIC_FLAG_NACK;
                            sig.the_flags&=~IIC_FLAG_READ;
                            sig.the_byte_loc=sig.the_bytes;
                            sig.the_rt_all_bytes=1;
                        }
                    }
                    else sig.the_retry_cnt=0;
                }
            }
            else { // now low, to high. C10 to C2. since C1 will not entered.
                SclHigh();
                if (isSclLow_W()){ // clock stretched by slave at bit and byte(ack) level
                    sig.the_proc=IIC_PROC_WAIT_STATES;
                    sig.the_next_proc=IIC_PROC_WRITE_DATA;
                    sig.the_flags|=IIC_FLAG_WAIT_CLK_H_DATA;
                    sig.the_tmo_cnt=IIC_TMO_WAIT_CLK_RW;
                }
                else sig.the_clk_flip=1;
                return;
            }
        }
        if (sig.the_byte_loc<sig.the_bytes){ // means there is data ready to send. note that wastes a cycle at every byte.
            sig.the_data=(*(sig.the_buff))[sig.the_byte_loc++];
            sig.the_bit_loc=10;
        }
        else if (!sig.the_rt_all_bytes){ // and data preparing wait is here
            // do the clk wait while cb-func, the (sig.the_data_cb)().
            // note that the cb function has 4 vars to modify:
            // the_rt_all_bytes to 1 if no more write preparing
            // sig.the_byte_loc to 0 for a new start index to the new/old buffer, or keep intact.
            // sig.the_bytes to the number of all bytes(corresponds to the_byte_loc) which is ready.
            // sig.the_buff if need to change buffer target.

            sig.the_flags|=IIC_FLAG_WAIT_DATA;
            sig.the_proc=IIC_PROC_WAIT_STATES;
            sig.the_next_proc=IIC_PROC_WRITE_DATA;
            sig.the_tmo_cnt=IIC_TMO_WAIT_RTDATA(3);
        }
        else { // terminated condition. @A1h.
            SclLow();
            if (sig.the_flags&IIC_FLAG_WAIT_DATA){ // the previous preparing read timeout.
                // so need to drop the succeeding read if have.
                sig.the_flags&=~IIC_FLAG_READ;
            }
            if (sig.the_flags&IIC_FLAG_READ){
                SdaHigh();
                sig.the_flags|=IIC_FLAG_SR_CONTINUE;
                sig.the_proc=IIC_PROC_REPEATSTART;
                sig.the_next_proc=IIC_PROC_ADDRESS;
            }
            else if (sig.the_queue->nxt && !(sig.the_flags&IIC_OPT_DIRECT_STOP)){
                SdaHigh();
                sig.the_flags|=IIC_FLAG_SR_NEW;
                sig.the_proc=IIC_PROC_REPEATSTART;
                sig.the_next_proc=IIC_PROC_ADDRESS;
            }
            else {
                SdaLow();
                if (!(sig.the_flags&IIC_FLAG_NACK)) sig.the_flags=IIC_FLAG_SUCCESS;
                sig.the_proc=IIC_PROC_STOP;
            }
        }
    };

    SRAM_ACES static void iicproc_stop(){ // so, here treats with low, low or high, high.
        // in(SCLD2h, SDAD2h, IIC_PROC_STOP). reset.
        // in(SCLB0l, SDAl) if nacked then to stop. address.
        // in(SCLB0l, SDAl) for stop. read.
        // in(SCLA0l, SDAl) for stop. write.

        if (sig.temp==1){
            if (sig.the_gate) return;

            if (sig.the_queue->obj->sr_p_cb_func) sig.the_queue->obj->sr_p_cb_func();

            sig.the_queue->Remove(&(sig.the_queue), &(sig.the_unqueue));
            sig.temp=0;
            sig.the_proc=IIC_PROC_IDLE;
        }
        else if (!sig.temp && isSclHigh()){SclLow_w(); SdaLow(); sig.temp=2;}
        else {
            SclHigh_w();
            SdaHigh_W();

                MYIICDEBUG(printf("\r\n<<<result flag: 0x%x>>>\r\n", sig.the_flags));

            sig.the_queue->obj->flags&=~IIC_FLAG_STATUS;
            sig.the_queue->obj->flags|=(sig.the_flags&IIC_FLAG_STATUS);
            sig.the_queue->obj->flags&=~IIC_FLAG_BUSY;
            if (sig.the_gate) sig.temp=1;
            else {

                if (sig.the_queue->obj->sr_p_cb_func) sig.the_queue->obj->sr_p_cb_func();

                sig.the_queue->Remove(&(sig.the_queue), &(sig.the_unqueue));
                sig.temp=0;
                sig.the_proc=IIC_PROC_IDLE;
            }
        }
    };

    void init_vars(){
        SRAM_ACES static void (*x[])()={
            0,
            cIIC::iicproc_start,
            cIIC::iicproc_repeatstart,
            cIIC::iicproc_retry,
            cIIC::iicproc_address,
            cIIC::iicproc_write_data,
            cIIC::iicproc_read_data,
            cIIC::iicproc_wait_states,
            cIIC::iicproc_stop,
            cIIC::iicproc_reset,
        };
        sig.the_procs=x; // multiple assignments but ok not frequently.
        master.Clear();
    };

    bool queue_check_exists(){
        sig.the_gate=(sIicQueue*)1; // lock all if 1.
        sIicQueue *q=sig.the_queue;
        while (q){
            if (q->obj==&master){sig.the_gate=0; return true;}
            q=q->nxt;
        }

        // gc recycled.
        delete sig.the_unqueue;
        sig.the_unqueue=0;

        sig.the_gate=0;
        return false;
    };

    public:

    SRAM_ACES static void isr_iicproc(){
        if (sig.the_proc){

                MYIICDEBUG(\
                    printf("\r\n[%d 0x%x %d] %d %s: SCL(%d), SDA(%d)\r\n",\
                    sig.the_proc,\
                    sig.the_flags,\
                    sig.temp+sig.the_retry_cnt,\
                    (unsigned)(sig.the_queue),\
                    dbg_iic_stage_name(sig.the_proc),\
                    IS_SCL_H,\
                    IS_SDA_H\
                ));

            sig.the_procs[sig.the_proc]();
        }
        else if (sig.the_queue && (sig.the_gate!=(sIicQueue*)1)){
            sig.the_proc=IIC_PROC_START;

                MYIICDEBUG(\
                    printf("\r\n[%d 0x%x %d] %d %s: SCL(%d), SDA(%d)\r\n",\
                    sig.the_proc,\
                    sig.the_flags,\
                    sig.temp+sig.the_retry_cnt,\
                    (unsigned)(sig.the_queue),\
                    dbg_iic_stage_name(sig.the_proc),\
                    IS_SCL_H,\
                    IS_SDA_H\
                ));

            sig.the_procs[sig.the_proc]();
        }
    };

    typedef enum eIIC_OPTS{
        IIC_OPT_NULL=               0,
        IIC_OPT_MASTER=             2,
        IIC_OPT_SLAVE=              ~2,
        IIC_OPT_WRITE=              4,
        IIC_OPT_READ=               8,
        IIC_OPT_W_THEN_R=           16,
        IIC_OPT_R_THEN_W=           32,
        IIC_OPT_RT_WRITE=           IIC_FLAG_RT_WRITE, // 0x100
        IIC_OPT_RT_READ=            IIC_FLAG_RT_READ, // 0x200
        IIC_OPT_RT_TRANSFER=        IIC_FLAG_RT_TRANSFER, // (0x300)
        IIC_OPT_DIRECT_STOP=        0x8000, // means directly stop by this request even succeeding new requests exist; no repeat start.
    } eIIC_OPTS;

    bool Config(unsigned addr_7b, unsigned n_r, unsigned n_w, unsigned char *ri_buff, unsigned char *wo_buff, eIIC_OPTS opt, bool rt=false){
        // all the parameters must defined(0 if no use). if realtime use, the RtConfig() must then called.
        // opt: IIC_OPT_WRITE/IIC_OPT_READ/IIC_OPT_W_THEN_R/IIC_OPT_R_THEN_W. the or bits are: IIC_OPT_DIRECT_STOP.

        if (queue_check_exists()) return false; // the submitted one is still processing.

        init_vars();

        switch (opt&~IIC_OPT_DIRECT_STOP){
            case IIC_OPT_WRITE:
                if (!n_w && !rt) return false;
                master.write_cnt=n_w;
                master.rt_obuff=wo_buff;
                master.flags=(IIC_FLAG_WRITE_FIRST|IIC_FLAG_WRITE);
                break;
            case IIC_OPT_READ:
                if (!n_r && !rt) return false;
                if (rt && !ri_buff && n_r) return false;
                master.read_cnt=n_r;
                master.rt_ibuff=ri_buff;
                master.flags=IIC_FLAG_READ;
                break;
            case IIC_OPT_W_THEN_R:
                if ((!n_w || !n_r) && !rt) return false;
                if (rt && !ri_buff && n_r) return false;
                master.read_cnt=n_r;
                master.write_cnt=n_w;
                master.rt_ibuff=ri_buff;
                master.rt_obuff=wo_buff;
                master.flags=(IIC_FLAG_WRITE_FIRST|IIC_FLAG_WRITE|IIC_FLAG_READ);
                break;
            case IIC_OPT_R_THEN_W:
                if ((!n_w || !n_r) && !rt) return false;
                if (rt && !ri_buff && n_r) return false;
                master.read_cnt=n_r;
                master.write_cnt=n_w;
                master.rt_ibuff=ri_buff;
                master.rt_obuff=wo_buff;
                master.flags=(IIC_FLAG_WRITE|IIC_FLAG_READ);
                break;
            default:
                return false;
        }

        master.address=(addr_7b<<1);
        master.flags|=(opt&IIC_OPT_DIRECT_STOP);

        if (rt) master.flags|=IIC_FLAG_RT_TRANSFER;
        else {
            sig.the_gate=(sIicQueue*)1; // lock all if 1.
            sig.the_queue->Add(&(sig.the_queue), &(this->master));
            sig.the_gate=0;
        }
        return true;
    };

    // for realtime process. option would be IIC_OPT_RT_WRITE/IIC_OPT_RT_READ/IIC_OPT_RT_TRANSFER(both).
    bool RtConfig(eIIC_OPTS opt, void (*wcb_fooptr)(), void (*rcb_fooptr)()){ // called after configed.
        if ((master.flags&IIC_FLAG_RT_TRANSFER)!=IIC_FLAG_RT_TRANSFER) return false;
        opt=(eIIC_OPTS)(opt&IIC_OPT_RT_TRANSFER);
        if (!opt) return false;
        else if ((opt&IIC_OPT_RT_WRITE) && !(master.flags&IIC_FLAG_WRITE)) return false;
        else if ((opt&IIC_OPT_RT_READ) && !(master.flags&IIC_FLAG_READ)) return false;
        master.rt_wdata_prepare=wcb_fooptr;
        master.rt_rdata_postacc=rcb_fooptr;
        master.flags&=~IIC_FLAG_RT_TRANSFER;
        master.flags|=opt;
        sig.the_gate=(sIicQueue*)1; // lock all if 1.
        sig.the_queue->Add(&(sig.the_queue), &(this->master));
        sig.the_gate=0;
        return true;
    };

    bool RepRun(){ // only for static mode. called after configed.
        if (!master.address) return false;
        if (master.flags&IIC_FLAG_RT_TRANSFER) return false;
        if (queue_check_exists()) return false;
        master.flags&=~IIC_FLAG_STATUS;
        sig.the_gate=(sIicQueue*)1; // lock all if 1.
        sig.the_queue->Add(&(sig.the_queue), &(this->master));
        sig.the_gate=0;
        return true;
    };

    bool SetCb(void (*start_cb)(), void (*stop_cb)()){ // called after configed. 0 if not used.
        if (!master.address) return false;
        master.sr_s_cb_func=start_cb;
        master.sr_p_cb_func=stop_cb;
        return true;
    };

    cIIC(){;};

    cIIC(unsigned addr_7b, unsigned n_r, unsigned n_w, unsigned char *ri_buff, unsigned char *wo_buff, eIIC_OPTS opt, bool rt=false){
        this->Config(addr_7b, n_r, n_w, ri_buff, wo_buff, opt, rt);
    };

    cIIC(unsigned addr_7b, unsigned n_r, unsigned char *ri_buff, bool rt=false){ // read
        this->Config(addr_7b, n_r, 0, ri_buff, 0, IIC_OPT_READ, rt);
    };

    cIIC(unsigned addr_7b, unsigned char *wo_buff, unsigned n_w, bool rt=false){ // write
        this->Config(addr_7b, 0, n_w, 0, wo_buff, IIC_OPT_WRITE, rt);
    };
};

SRAM_ACES cIIC::sIicMasterSig cIIC::sig; // sig must be inited before all iic objects




    SRAM_ACES void cb_test(){;}
    unsigned char mybuf[100];

    cIIC a(0, 5, mybuf); // read
    cIIC b(0, mybuf, 5); // write

    cIIC art(0, 5, mybuf, true); // read rt
    //art.RtConfig(cIIC::IIC_OPT_RT_READ, cb_test, cb_test); // read rt

    cIIC brt(0, mybuf, 5, true); // write rt
    //brt.RtConfig(cIIC::IIC_OPT_RT_WRITE, cb_test, cb_test); // write rt

    cIIC d(0, 5, 5, mybuf, mybuf, cIIC::IIC_OPT_R_THEN_W); // rw
    cIIC e(0, 5, 5, mybuf, mybuf, cIIC::IIC_OPT_W_THEN_R); // wr

    cIIC drt(0, 5, 5, mybuf, mybuf, cIIC::IIC_OPT_R_THEN_W, true); // rw rt
    //drt.RtConfig(cIIC::IIC_OPT_RT_TRANSFER, cb_test, cb_test); // rw rt

    cIIC ert(0, 5, 5, mybuf, mybuf, cIIC::IIC_OPT_W_THEN_R, true); // wr rt
    //ert.RtConfig(cIIC::IIC_OPT_RT_TRANSFER, cb_test, cb_test); // wr rt




void setup(){
    Serial.begin(115200);
    while (!Serial) delay(10);




    //cIIC a(0, 5, mybuf); // read
    //cIIC b(0, mybuf, 5); // write

    //cIIC art(0, 5, mybuf, true); // read rt
    art.RtConfig(cIIC::IIC_OPT_RT_READ, cb_test, cb_test); // read rt

    //cIIC brt(0, mybuf, 5, true); // write rt
    brt.RtConfig(cIIC::IIC_OPT_RT_WRITE, cb_test, cb_test); // write rt

    //cIIC d(0, 5, 5, mybuf, mybuf, cIIC::IIC_OPT_R_THEN_W); // rw
    //cIIC e(0, 5, 5, mybuf, mybuf, cIIC::IIC_OPT_W_THEN_R); // wr

    //cIIC drt(0, 5, 5, mybuf, mybuf, cIIC::IIC_OPT_R_THEN_W, true); // rw rt
    drt.RtConfig(cIIC::IIC_OPT_RT_TRANSFER, cb_test, cb_test); // rw rt

    //cIIC ert(0, 5, 5, mybuf, mybuf, cIIC::IIC_OPT_W_THEN_R, true); // wr rt
    ert.RtConfig(cIIC::IIC_OPT_RT_TRANSFER, cb_test, cb_test); // wr rt
}

void loop(){
    cIIC::isr_iicproc();
}
